/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.List;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.ProvisionInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;

import com.google.common.collect.ClassToInstanceMap;
import com.google.inject.Binding;
import com.google.inject.Key;
import com.google.inject.Provider;
import com.google.inject.TypeLiteral;
import com.google.inject.spi.ElementSource;

/**
 * This {@link MethodInterceptor} is intended to be used with {@link DecoderModule}s where multiple
 * ones "produce" the same {@link PhenotypeDecoder}.
 * It intercepts the decode method call (defined by an annotation) and launches all sub-
 * {@link DecoderModule}s before executing its own decoding. Sub-{@link DecoderModule}s are not
 * allowed to interfere with each other.
 * 
 * @author diewald
 */
public class DecodeInterceptor extends ProvisionInterceptor<Phenotype> {

	/** Executes sub-decoders of a {@code @PluginDecoder} annotated decode method. */
	@Override
	protected Phenotype executeSubOperations(MethodInvocation invocation,
			Class<Phenotype> returnType) throws Exception {
		Phenotype phenotype = null;
		List<Binding<Phenotype>> subDecoderBindings =
				injector.findBindingsByType(TypeLiteral.get(returnType));

		for(Binding<Phenotype> decodeBinding : subDecoderBindings) {
			Provider<Phenotype> provider = decodeBinding.getProvider();
			ElementSource source = (ElementSource)decodeBinding.getSource();
			Object declaringSource = source.getDeclaringSource();
			if(!(declaringSource instanceof Method)) {
				throw new Exception("An element providing the type " + returnType.getSimpleName() +
						" is not a method. This is not allowed in this context: This" +
						" interceptor is only valid for decode methods (annotated with" +
						" @Decodes).");
			}
			Method declaringMethod = (Method)declaringSource;
			if(!overrides(invocation.getMethod(), declaringMethod)) {
				// The assignment is barely needed, since "sub-decoders" typically update the
				// Phenotype by its input parameter (one input parameter equals the output).
				phenotype = provider.get();
				phenotype.setSubDecoding(true);
			}
		}

		if(phenotype != null) {
			// phenotype == null ==> No sub-decoders were launched.
			phenotype.setSubDecoding(false);
		}
		return phenotype;
	}

	/**
	 * Returns true if {@code override} overrides {@code base}, assumes that the signatures match.
	 * Taken from Google Guice.
	 */
	private static boolean overrides(Method override, Method base) {
		// First, check the inheritance hierarchy, then compare the method modifiers and signatures.
		if(!base.getDeclaringClass().isAssignableFrom(override.getDeclaringClass())) {
			return false;
		}

		// See JLS section 8.4.8.1
		int modifiers = base.getModifiers();
		if(Modifier.isPublic(modifiers) || Modifier.isProtected(modifiers)) {
			return true;
		}
		if(Modifier.isPrivate(modifiers)) {
			return false;
		}
		// b must be package-private
		return override.getDeclaringClass().getPackage()
				.equals(base.getDeclaringClass().getPackage());
	}

	/** {@inheritDoc} */
	@Override
	protected Key<ClassToInstanceMap<Phenotype>> createInstanceHolderIdentifier() {
		return Key.get(new TypeLiteral<ClassToInstanceMap<Phenotype>>() {/**/
		});
	}
}
