/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.EncodingProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ProvidingGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.start.Opt4JTask;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Provides;

/**
 * {@link Decoder} that handles dependencies between {@link Decoder}s, {@link Genotype}s, and
 * {@link Phenotype}s using the Guice framework.
 * 
 * @author diewald
 */
public class GuiceDecoder extends AbstractModule implements
		Decoder<ProvidingGenotype, ProvidingPhenotype> {

	/** Defines the desired exploration result type. */
	// TODO: Remove this hard coded line and get the desired result from the backend interface.
	protected Set<Class<? extends Phenotype>> resultTypes;

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<DecoderModule<?>> enabledDecoders;

	/** {@link Opt4JTask} provider used to obtain the opt4j {@link Injector}. */
	@Inject
	protected Opt4JDseTaskProvider opt4JProvider;

	/** Child {@link Injector} for the dependency management of the decode phase. */
	protected Injector decodingInjector;

	/**
	 * Holds the {@link MethodInterceptor} that is executed when a decode method capable of
	 * launching sub-decoders is hit.
	 */
	protected DecodeInterceptor decodingInterceptor;

	/** Container for decoded {@link Phenotype}s. */
	protected ClassToInstanceMap<Phenotype> aggregatePhenotypes = MutableClassToInstanceMap
			.create();

	/**
	 * Allows to construct prototype providers. Avoids triggering the code generation by ByteBuddy
	 * multiple times.
	 */
	protected EncodingProviderModule<Phenotype> protoProviderModule;

	// protected DecodingInjectorProvider decodingInjectorProvider = new
	// DecodingInjectorProvider();
	/** Reference to the input {@link Genotype}. */
	// @ThreadScoped
	protected ThreadLocal<CompositeGenotype<Class<? extends Genotype>, ? extends Genotype>> genotype =
			new ThreadLocal<>();

	/** Constructor. Remembers the demanded {@link Phenotype}ss to be produced. */
	public GuiceDecoder(Set<Class<? extends Phenotype>> resultTypes,
			Collection<DecoderModule<?>> enabledDecoders) {
		this.resultTypes = resultTypes;
		this.enabledDecoders = enabledDecoders;
	}

	// To be used later: Constructor injection is favorable.
	// /** Constructor. Gets the Opt4J injector by Guice. */
	// @Inject
	// public GuiceDecoder(Opt4JDseTaskProvider provider) {
	// decodingInjector = provider.get().get().createChildInjector(this);
	// }

	/** {@inheritDoc} */
	@Override
	public ProvidingPhenotype decode(ProvidingGenotype genotype) {
		// TODO(AD): See configure(): We need a service mechanism and user configuration to select
		// the desired set of activated decoders.
		decodingInjector =
				opt4JProvider.get().get().createChildInjector(this, genotype.getProviderModule());
		DecodeInterceptor interceptor = decodingInjector.getInstance(DecodeInterceptor.class);
		interceptor.setInterceptParameters(decodingInjector, Phenotype.class);
		// decodingInjectorProvider.setDecodingInjector(decodingInjector);

		// FIXME: Hack to get the correct instance: Is multiprovisioning supported by Guice?
		Class<? extends Phenotype> resultType = getFirst(resultTypes).get();
		decodingInjector.getInstance(resultType);

		// The Provider Prototypes must be installed here, since the phenotypes to be generated are
		// only know here, not in the constructor. Once the meta-analysis of dependecies is
		// completely frontloaded, the prototype providers could also be installed in the
		// constructor.
		// NOTE: This provider module may NEVER be installed in the Guice configuration section, as
		// it would produce duplicate @Provides methods. This module is intended to capture decoding
		// results for the evaluation phase.
		createProtoProvider();
		EncodingProviderModule<Phenotype> phenotProviderModule =
				new EncodingProviderGenerator<Phenotype>().generateInstanceProviders(
						protoProviderModule, aggregatePhenotypes.values());

		ClassToInstanceMap<Phenotype> solutionHolder = MutableClassToInstanceMap.create();
		solutionHolder.putAll(aggregatePhenotypes);
		ProvidingPhenotype providingPhenotype =
				new ProvidingPhenotype(genotype.getProviderModule(), phenotProviderModule,
						solutionHolder, genotype.getGenotypeMap());
		aggregatePhenotypes.clear();

		return providingPhenotype;
	}

	/** Provides the composite {@link Genotype} as required by the Guice framework. */
	@Provides
	// @ThreadScoped
			public
			CompositeGenotype<Class<? extends Genotype>, ? extends Genotype> getGenotype() {
		return genotype.get();
	}

	/** Returns the sub-genotype of the given type, or {@code null} if none is found. */
	public <T extends Genotype> T getGenotype(Class<T> genotypeType) {
		return getGenotype().get(genotypeType);
	}

	/** Returns the Container map of already decoded {@link Phenotype}s. */
	@Provides
	public ClassToInstanceMap<Phenotype> getAggregatePhenos() {
		return aggregatePhenotypes;
	}

	/**
	 * Constructs a prototype {@link EncodingProviderModule}. This method shall be executed only
	 * once, since it is required only for provider instance construction. It is synchronized since
	 * it may be launched from different decode threads.
	 */
	private synchronized void createProtoProvider() {
		if(protoProviderModule != null) {
			return;
		}
		this.protoProviderModule =
				new EncodingProviderGenerator<Phenotype>().generatePrototypes(aggregatePhenotypes
						.keySet());
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		// bindScope(ThreadScoped.class, new ThreadScope());
		install(this);
		// install(decodingInjectorProvider);

		for(DecoderModule<?> module : enabledDecoders) {
			install(module);
		}
	}
}
