/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.pickInstanceOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMultiMap;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.opt4j.core.Genotype;
import org.opt4j.core.Objective;
import org.opt4j.core.Objectives;
import org.opt4j.core.Value;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.MutableClassToInstanceMap;

/**
 * Base class for Evaluators of {@link Phenotype}s.
 * 
 * @author diewald
 */
public abstract class CompositeEvaluatorBase<P extends Phenotype> implements CompositeEvaluator<P> {
	/** Flag to mark that marks the top-most evaluator in the composite structure. */
	protected boolean isTopLevelEvaluator = false;

	/** Flag to determine if the {@link #init()} routine was launched for this instance. */
	private boolean isInitialized = false;

	/** References the models provided by the {@link IExplorationTransformationService}. */
	protected ClassToInstanceMap<EObject> transformedModels = MutableClassToInstanceMap.create();

	/** Collection of required models demanded by the evaluators. */
	protected Collection<Class<? extends EObject>> requiredTransformedModelTypes;

	/**
	 * Correlates the {@link Phenotype}s with the corresponding {@link ComposableEvaluator}s forming
	 * this CompositeEvaluator.
	 */
	private SubClassToInstanceMultiMap<Phenotype, ComposableEvaluator<? extends Phenotype>> subEvaluators =
			new SubClassToInstanceMultiMap<>(Phenotype.class);

	/** Constructor. */
	public CompositeEvaluatorBase() {
		if(!ComposableEvaluator.class.isAssignableFrom(getClass())) {
			isTopLevelEvaluator = true;
		}
	}

	/**
	 * Performs the initialization of {@code this} {@link CompositeEvaluator} that cannot be done in
	 * the constructor, e.g., due to the missing registration of sub evaluators.
	 * It is executed when {@link #evaluateInternal(Phenotype)} is executed for the first
	 * time of this instance.
	 */
	protected void init() {
		if(isInitialized) {
			return;
		}
		isInitialized = true;

		if(isTopLevelEvaluator) {
			// Set<IExplorationTargetEvaluator<?, ?, ?>> allSubEvaluators = new HashSet<>();
			// getAllSubEvaluators().forEach(
			// e -> allSubEvaluators.add(((EvaluatorWrapper<?, ?, ?>)e).getEvaluator()));
			//
			// @SuppressWarnings("unchecked") Collection<Class<? extends EObject>> tempReqModels =
			// getService(IExplorationEvaluatationService.class).getRequiredTransformedModels(
			// allSubEvaluators);
			// requiredTransformedModelTypes = tempReqModels;
		}
	}

	/**
	 * Internal method that calls the evaluators for {@link ComposableGenotype}s.
	 * <b>NOTE:</b> Sub-implementations must call {@code super.decode(P)}!!!
	 * 
	 * @param phenotype
	 *            {@link Phenotype} that shall be evaluated w.r.t. to the
	 *            {@link ExplorationSpecification}.
	 * @return Evaluation result as {@link Objectives}.
	 */
	public <S extends Phenotype> Objectives evaluateInternal(P phenotype) {
		init();

		Multimap<Objective, Value<?>> subObjectivesEvaluations = HashMultimap.create();
		Multimap<Objective, S> objectiveContribPheno = HashMultimap.create();

		FlatPhenotypeMap<Phenotype> phenotypeTypeMap = null;
		if(isTopLevelEvaluator) {
			phenotypeTypeMap = constructPhenotypeMap(phenotype);
		} else {
			// TODO: Create an EvaluationException and use it here.
			throw new RuntimeException(
					"The phenotype type map could not be created for the top-most evaluator.");
		}

		ClassToInstanceMap<EObject> transformedModels = MutableClassToInstanceMap.create();
		if(isTopLevelEvaluator && !requiredTransformedModelTypes.isEmpty()) {
			try {
				// TODO: Readd the transformation service to the evaluators.
				// transformedModels =
				// getService(IExplorationTransformationService.class).getTransformedModels(
				// DSE_INTERNAL, null, phenotypeTypeMap.asExplorationEncoding(),
				// systemModelAdapter.getInputModels(), requiredTransformedModelTypes,
				// null);
			} catch(Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		Iterator<Entry<Class<? extends Phenotype>, Phenotype>> phenoIter = phenotype.iterator();
		while(phenoIter.hasNext()) {
			Entry<Class<? extends Phenotype>, Phenotype> phenoEntry = phenoIter.next();
			@SuppressWarnings("unchecked") S subPhenotype = (S)phenoEntry.getValue();
			@SuppressWarnings("unchecked") Class<S> key = (Class<S>)phenoEntry.getKey();
			Collection<ComposableEvaluator<? extends Phenotype>> phenoSubEvaluators =
					subEvaluators.get(phenoEntry.getValue().getClass());
			if(phenoSubEvaluators != null) {
				// If no subEvaluator is registered, only decoding shall be performed -> no
				// evaluation.
				Objectives evalResults = new Objectives();
				for(ComposableEvaluator<? extends Phenotype> subEvaluator : phenoSubEvaluators) {
					evalResults =
							subEvaluator.evaluateComposable(key, subPhenotype, phenotypeTypeMap,
									transformedModels);
					Iterator<Entry<Objective, Value<?>>> resultIterator = evalResults.iterator();
					while(resultIterator.hasNext()) {
						Entry<Objective, Value<?>> result = resultIterator.next();
						objectiveContribPheno.put(result.getKey(), subPhenotype);
						subObjectivesEvaluations.put(result.getKey(), result.getValue());
					}
				}
			}
		}

		transformedModels.clear();

		Objectives mergedObjectives = new Objectives();
		for(Entry<Objective, Collection<S>> objPheno : objectiveContribPheno.asMap().entrySet()) {
			// TODO : re-add merged value as a separate value.
			Value<?> objectiveEvalValue =
					mergeSubEvaluations(subObjectivesEvaluations.get(objPheno.getKey()));
			mergedObjectives.add(objPheno.getKey(), objectiveEvalValue);
			// Objective objective = objPheno.getKey();
			// for(Value<?> val
			// mergedObjectives.add(objPheno.getKey(),
			// subObjectivesEvaluations.get(objPheno.getKey()));
		}

		return mergedObjectives;
	}

	// /** {@inheritDoc} */
	// @Override
	// public Objectives evaluate(P phenotype) {
	// return evaluateInternal(phenotype);
	// }
	//
	// /** {@inheritDoc} */
	// @Override
	// public <S extends Phenotype> void registerSubEvaluator(Class<S> subPhenotype,
	// ComposableEvaluator<S> subEvaluator) throws Exception {
	// subEvaluators.put(subPhenotype, subEvaluator);
	// }

	/** {@inheritDoc} */
	@Override
	public Collection<ComposableEvaluator<? extends Phenotype>> getSubEvaluators() {
		Collection<ComposableEvaluator<? extends Phenotype>> allSubEvaluators = new HashSet<>();
		for(ComposableEvaluator<? extends Phenotype> evaluators : subEvaluators.values()) {
			allSubEvaluators.add(evaluators);
		}
		return allSubEvaluators;
	}

	/** {@inheritDoc} */
	@Override
	public Set<ComposableEvaluator<? extends Phenotype>> getAllSubEvaluators() {
		Set<ComposableEvaluator<? extends Phenotype>> allSubEvaluators = new HashSet<>();
		// Get the direct sub-Evaluators.
		allSubEvaluators.addAll(subEvaluators.values());

		// Recursively fetch the nested sub-Evaluators.
		@SuppressWarnings("rawtypes") Collection<CompositeEvaluator> compositeSubEvaluators =
				pickInstanceOf(CompositeEvaluator.class, subEvaluators.values());
		for(CompositeEvaluator<?> curCompositeEval : compositeSubEvaluators) {
			allSubEvaluators.addAll(curCompositeEval.getAllSubEvaluators());
		}

		return allSubEvaluators;
	}

	/**
	 * Constructs a flattened version of the given {@link Phenotype}, i.e. a
	 * {@link FlatPhenotypeMap}.
	 */
	private <G extends Genotype, CP extends Phenotype, SP extends Phenotype>
			FlatPhenotypeMap<Phenotype> constructPhenotypeMap(CP phenotype) {
		FlatPhenotypeMap<Phenotype> phenotypeTypeMap = new FlatPhenotypeMap<>();
		@SuppressWarnings("unchecked") LinkedList<SP> subPhenotypes =
				new LinkedList<SP>((Collection<? extends SP>)phenotype.getSubPhenotypes());
		while(!subPhenotypes.isEmpty()) {
			SP currSubPhenotype = subPhenotypes.poll();
			@SuppressWarnings("unchecked") Class<SP> currSupPhenotypeType =
					(Class<SP>)currSubPhenotype.getClass();
			phenotypeTypeMap.putEncoding(currSupPhenotypeType, currSubPhenotype);
			if(Phenotype.class.isAssignableFrom(currSubPhenotype.getClass())) {
				@SuppressWarnings("unchecked") List<SP> subPhenos =
						(List<SP>)((Phenotype)currSubPhenotype).getSubPhenotypes();
				subPhenotypes.addAll(0, subPhenos);
			}
		}

		return phenotypeTypeMap;
	}
}
