/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.abs;
import static java.lang.Math.expm1;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Optional;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.SameSilInPartitionConstraint;
import org.fortiss.af3.safety.model.SIL;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Evaluates whether all {@link ITaskAdapter}s mapped to {@link Partition}s have the same
 * SIL/ASIL/DAL value and calculates a penality based on the difference in the Integrity Levels of
 * the tasks.
 * 
 * @author diewald
 */
public class SameSilInPartitionConstraintEvaluator
		extends
		EvaluatorModule<Double, SameSilInPartitionConstraint, TaskMappingEncoding<ITaskAdapter<?>, InstantiatedTaskMappingEntry>>
		implements
		IConstraintEvaluator<Double, SameSilInPartitionConstraint, TaskMappingEncoding<ITaskAdapter<?>, InstantiatedTaskMappingEntry>> {

	/** Prototype Constructor. */
	public SameSilInPartitionConstraintEvaluator(
			Class<SameSilInPartitionConstraint> explorationTargetType) {
		super(explorationTargetType);
	}

	/** Constructor. */
	public SameSilInPartitionConstraintEvaluator(SameSilInPartitionConstraint explorationTarget) {
		super(explorationTarget);
	}

	/**
	 * Penalty function that ensures only {@link ITaskAdapter} with the same {@link SIL} reside in
	 * the same {@link Partition}. For each {@link Partition}, an exponential penalty is calculated.
	 * The resulting values are summed to provide an estimate for the overall system.
	 */
	@Provides
	@Named("SameSilInPartitionConstraintEvaluator")
	public DoubleValue evaluateGoal(FailSilentTaskMappingEncoding phenotype) {
		double resVal = 0.0;
		for(Partition partition : phenotype.getResources()) {
			// FIXME: remove casts by specifying the TaskMapping.
			Collection<ITaskAdapter<?>> allocatedTasks = phenotype.getRequesters(partition);
			Optional<ITaskAdapter<?>> anyTask = getFirst(allocatedTasks);
			if(!anyTask.isPresent()) {
				continue;
			}

			int derivationSum = 0;
			int firstSilVal = anyTask.get().getRequiredSafetyIntegrityLevel().getValue();
			for(ITaskAdapter<?> task : allocatedTasks) {
				int currentSilVal = task.getRequiredSafetyIntegrityLevel().getValue();
				derivationSum += abs(firstSilVal - currentSilVal);
			}

			// TODO: Will the evaluator perform more efficiently if the penalty function is
			// calculated over the sum of violation of partitions, or (as it is now) a sum over the
			// penalties per partition mismatch.
			resVal += expm1(derivationSum);
		}
		return new DoubleValue(resVal);
	}

	/** {@inheritDoc} */
	@Override
	public boolean validate(ExplorationConstraint<?> expTarget,
			TaskMappingEncoding<ITaskAdapter<?>, InstantiatedTaskMappingEntry> solution,
			Value<?> value) {
		return value.getDouble() <= 10e-6;
	}

}
