/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.random;

import java.lang.reflect.Constructor;
import java.security.SecureRandom;
import java.util.Random;

import com.google.common.primitives.Longs;

/**
 * Central random number generator for the Exploration. It provides access to both
 * deterministic/pseudo-random and "real" random number generators. The type of generator must be
 * defined during construction time.
 * <p>
 * The driving idea behind this centralized random number generator is to provide the possibility to
 * generate reproducible results by providing identical seeds. On the other hand "real" randomness
 * must be also provided for practical applications.
 * 
 * @author diewald
 */
public class RandomExploration {

	/** Defines the types of random number generators that are used within the DSE. */
	public static enum RandomnessType {
		/** Deterministic random number generator (using the provided seed). */
		deterministic,
		/** Pseudo random number generator (ignoring the seed). */
		pseudoRandom,
		/** "Real" random number generator. */
		realRandom;
	}

	/** Defines the type of random numbers that this class shall provide. */
	RandomnessType randType;

	/** Thread-safe singleton. */
	private static final class InstanceHolder {
		/** The actual instance of this single. Synchronization is done by the ClassLoader. */
		static final RandomExploration INSTANCE = new RandomExploration();
	}

	/** Constructor. Avoid instantiation by other objects. */
	private RandomExploration() {
	}

	/** Provides the RandomExploration number generator singleton. */
	public static RandomExploration getInstance() {
		return InstanceHolder.INSTANCE;
	}

	/** Flag to allow the singleton to be initialized only once. */
	private boolean isInitialized = false;

	/** Actual random number Generator. */
	private Random randGenerator;

	/**
	 * Initialize the central random number generator of the DSE by providing information about the
	 * randomness type and the seed that shall be used for this exploration run.
	 * 
	 * @param randType
	 *            {@link RandomnessType} that shall be used within the DSE.
	 * @param seed
	 *            Seed to be forwarded to the "real" random number generator.
	 * @throws Exception
	 *             if the centralized random number generator has been initialized before.
	 */
	public void init(RandomnessType randType, long seed) throws Exception {
		if(isInitialized) {
			throw new Exception("The RandomExploration singleton can only be initialized once.");
		}
		if(randType == RandomnessType.deterministic) {
			randGenerator = new Random(seed);
		} else if(randType == RandomnessType.pseudoRandom) {
			randGenerator = new Random();
		} else if(randType == RandomnessType.realRandom) {
			randGenerator = new SecureRandom(Longs.toByteArray(seed));
		}
		isInitialized = true;
	}

	/**
	 * Re-Initialize the central random number generator of the DSE, which must be performed if
	 * successive runs of the DSE shall have a deterministic behavior.
	 * 
	 * @throws Exception
	 *             Exception if the singleton has not been initialized before, or if another
	 *             instance of the backend random number generator cannot be instantiated.
	 */
	public void reInit() throws Exception {
		if(!isInitialized) {
			throw new Exception(
					"The RandomExploration singleton can only be reinitialized if init(...) has been called before.");
		}
		Constructor<? extends Random> randomGenCtor =
				randGenerator.getClass().getDeclaredConstructor();
		randGenerator = randomGenCtor.newInstance();
	}

	/**
	 * Provides the next boolean value as given by the chosen random generator. Wrapper for
	 * {@link Random#nextBoolean()}.
	 */
	public boolean nextBoolean() {
		assert (randGenerator != null);

		return randGenerator.nextBoolean();
	}

	/**
	 * Provides the next double value as given by the chosen random generator. Wrapper for
	 * {@link Random#nextDouble()}.
	 */
	public double nextDouble() {
		assert (randGenerator != null);

		return randGenerator.nextDouble();
	}

	/**
	 * Provides the next boolean value as given by the chosen random generator. Wrapper for
	 * {@link Random#nextInt(int)}.
	 */
	public int nextInt(int n) {
		assert (randGenerator != null);

		return randGenerator.nextInt(n);
	}

	/** Returns whether the centralized random number generator has been initialized. */
	public boolean isInitialized() {
		return isInitialized;
	}
}
