/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature;

import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.optimizer.Operator;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;
import org.opt4j.operators.OperatorModule;

import com.google.inject.AbstractModule;

/**
 * 
 * Exploration modules can be categorized into three classes: Variable definitions, Problem
 * statements and transformations, and Evaluators (solution metrics). The variable definitions
 * consist of a {@link Genotype}s (variables), {@link Operator}s (modify variables), and
 * {@link Creator}s (variable initialization). Problem statements transform {@link Genotype}s into
 * evaluable solutions implemented by {@link Decoder}s that produce {@link Phenotype}s.
 * {@link Evaluator}s
 * rate {@link Phenotype}s such that the quality of a solution can be quantified.
 * 
 * The complete DSE process is based on DI: Creators, Operators, Decoders, and Evaluators have to be
 * implemented as Guice {@link AbstractModule}s. Therefore, the package
 * {@code org.fortiss.af3.exploration.alg.dse.module} contains base classes simplifying the
 * definition of exploration modules.
 * 
 * @author diewald
 */
public interface IExplorationFeature {

	/** Returns the name of the {@link IExplorationFeature}. */
	String getName();

	/**
	 * Returns the set of {@link Genotype}s that can be created by the registered
	 * {@link CreatorModule}s.
	 */
	Set<Class<? extends Genotype>> getCreatableGenotypes();

	/** Returns the set of registered {@link Creator}s. */
	Set<CreatorModule<?>> getCreators();

	/** Returns the set of registered {@link Operator}s. */
	Set<OperatorModule<?>> getOperators();

	/** Returns the set of registered {@link Decoder}s. */
	Set<DecoderModule<?>> getDecoders();

	/** Returns the set of registered {@link Evaluator}s. */
	Set<EvaluatorModule<?, ?, ?>> getEvaluators();
}
