/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.ExplorationAlgActivator;
import org.fortiss.af3.exploration.alg.exception.ExplorationServiceException;
import org.osgi.framework.BundleContext;
import org.osgi.framework.FrameworkUtil;

/**
 * The {@link ExplorationServiceManager} singleton manages all services offered by the
 * org.fortiss.af3.exploration.alg plugin.
 * 
 * Service instances are retrieved by calling {@link #getService(Class)} with the type of the
 * desired {@link IExplorationService}. If the {@link IExplorationService} has not been instantiated
 * yet, this step is automatically performed by {@code this} {@link ExplorationServiceManager} manager.
 * 
 * <p>
 * 
 * <b>NOTE:<b> Only one service instance can be registered for one service type. If it is desired to
 * exchange the service implementations during runtime,
 * {@link #replaceService(Class, IExplorationService)} must be used.
 * 
 * @author diewald
 */
public class ExplorationServiceManager {

	/** Holds the instance of the transformation service. */
	private volatile static ExplorationServiceManager INSTANCE;

	/** Associates {@link IExplorationService} types with their singleton instances. */
	private Map<Class<? extends IExplorationService>, IExplorationService> services =
			new HashMap<>();

	/** Constructor. */
	private ExplorationServiceManager() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationServiceManager getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationServiceManager();
		}
		return INSTANCE;
	}

	/**
	 * Returns the service instance that is associated with the given {@code serviceType}.
	 * 
	 * @param serviceType
	 *            Type of the service.
	 * @return The service's instance registered for the given service type.
	 */
	public static <T extends IExplorationService> T getService(Class<T> serviceType) {
		@SuppressWarnings("unchecked") T service = (T)getInstance().services.get(serviceType);
		if(service == null) {
			BundleContext ctx =
					FrameworkUtil.getBundle(ExplorationAlgActivator.class).getBundleContext();
			service = ctx.getService(ctx.getServiceReference(serviceType));
			getInstance().services.put(serviceType, service);
		}
		return service;
	}

	/**
	 * Replaces the association of the given {@code serviceType} with its existing service instance
	 * with the given instance. If no association exists already, the <i>new</i> association is
	 * registered anyways.
	 * 
	 * @param serviceType
	 *            Type of the service.
	 * @param serviceInstance
	 *            The sevice's replacement instance.
	 * @return The previously registered service instance, or {@code null} if none exists.
	 * @throws ExplorationServiceException
	 *             If the OSGI service holder throws an exception.
	 */
	public static <T extends IExplorationService, S extends T, U extends T> S replaceService(
			Class<T> serviceType, U serviceInstance) throws ExplorationServiceException {
		@SuppressWarnings("unchecked") S oldInstance = (S)getInstance().services.get(serviceType);
		// Decrease the service use by ungetting again.
		BundleContext ctx =
				FrameworkUtil.getBundle(ExplorationAlgActivator.class).getBundleContext();
		ctx.ungetService(ctx.getServiceReference(oldInstance.getClass()));

		getInstance().services.put(serviceType, serviceInstance);
		return oldInstance;
	}
}
