/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static com.google.common.collect.Sets.newHashSet;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.tooling.common.util.LambdaUtils.filterStream;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.exception.ExplorationServiceRuntimeException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.operators.OperatorModule;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.AbstractModule;
import com.google.inject.Module;

/**
 * Implementation of the {@link IExplorationFeatureService}.
 * 
 * @author diewald
 */
public class ExplorationFeatureService implements IExplorationFeatureService {

	/** Holds the registered {@link IExplorationFeature}s. */
	private ClassToInstanceMap<IExplorationFeature> explorationFeatures = MutableClassToInstanceMap
			.create();

	/**
	 * References all registered exploration modules of {@link IExplorationFeature}s for consistency
	 * checks. Differing between {@link Creator}s, {@link Decoder}s etc. is currently not needed.
	 */
	private ClassToInstanceMap<Module> registeredExpModules = MutableClassToInstanceMap.create();

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationFeatureService INSTANCE;

	/** Constructor. */
	private ExplorationFeatureService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationFeatureService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationFeatureService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Typically used to read extension point configs. Not needed since we won't use eclipse
		// extension points in the DSE any more.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Not needed.
	}

	/** {@inheritDoc} */
	@Override
	public void registerExplorationFeature(IExplorationFeature feature) {
		// Consistency checks.
		for(CreatorModule<?> creatorModule : feature.getCreators()) {
			Class<? extends AbstractModule> creatorClass = creatorModule.getClass();
			if(registeredExpModules.containsKey(creatorClass)) {
				throw new ExplorationServiceRuntimeException(getClass(), "The module " +
						creatorClass.getSimpleName() + " has already been registered!");
			}
			registeredExpModules.put(creatorClass, creatorModule);
		}
		for(DecoderModule<?> decoderModule : feature.getDecoders()) {
			Class<? extends AbstractModule> decoderClass = decoderModule.getClass();
			if(registeredExpModules.containsKey(decoderClass)) {
				throw new ExplorationServiceRuntimeException(getClass(), "The module " +
						decoderClass.getSimpleName() + " has already been registered!");
			}
			registeredExpModules.put(decoderClass, decoderModule);
		}
		for(OperatorModule<?> operatorModule : feature.getOperators()) {
			Class<? extends AbstractModule> operatorClass = operatorModule.getClass();
			if(registeredExpModules.containsKey(operatorClass)) {
				throw new ExplorationServiceRuntimeException(getClass(), "The module " +
						operatorClass.getSimpleName() + " has already been registered!");
			}
			registeredExpModules.put(operatorClass, operatorModule);
		}
		for(EvaluatorModule<?, ?, ?> evaluatorModule : feature.getEvaluators()) {
			getService(IExplorationEvaluatationService.class).registerEvaluator(evaluatorModule);
		}

		explorationFeatures.put(feature.getClass(), feature);
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IExplorationFeature> getExplorationFeatures() {
		return explorationFeatures.values();
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends Genotype>> getCreatableGenotypesOf(
			Collection<IExplorationFeature> features) {
		Set<Class<? extends Genotype>> creatableGenotypes = newHashSet();
		features.forEach(f -> creatableGenotypes.addAll(f.getCreatableGenotypes()));
		return creatableGenotypes;
	}

	/** {@inheritDoc} */
	@Override
	public void registerEvaluatorModule(IEvaluatorModule<?, ?, ?> evaluatorModule) {
		Class<? extends Module> evaluatorClass = evaluatorModule.getClass();
		if(registeredExpModules.containsKey(evaluatorClass)) {
			throw new ExplorationServiceRuntimeException(getClass(), "The module " +
					evaluatorClass.getSimpleName() + " has already been registered!");
		}
		registeredExpModules.put(evaluatorClass, evaluatorModule);
	}

	/** {@inheritDoc} */
	@Override
	public Set<IEvaluatorModule<?, ?, ?>> getEvaluatorModules() {
		Set<IEvaluatorModule<?, ?, ?>> evaluatorModules = newHashSet();
		filterStream(registeredExpModules.values(), e -> e instanceof EvaluatorModule).forEach(
				e -> evaluatorModules.add((EvaluatorModule<?, ?, ?>)e));
		return evaluatorModules;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IExplorationFeature> getExplorationFeatures(
			Collection<Class<? extends IExplorationFeature>> features) {
		Collection<IExplorationFeature> featureInstances = new HashSet<>();
		for(Class<? extends IExplorationFeature> featureType : features) {
			IExplorationFeature featureInstance = explorationFeatures.getInstance(featureType);
			if(featureInstance == null) {
				throw new ExplorationServiceRuntimeException(getClass(),
						"No exploration feature of the type " + featureType.getClass() +
								" has been registered.");
			}
			featureInstances.add(explorationFeatures.getInstance(featureType));
		}
		return featureInstances;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		// KEEP DESCRIPTION IN SYNC WITH THE SERVICE INTERFACE CLASS COMMENT!!!!
		return "Service for adding Exploration Modules to the DSE. The define the concrete problems"
				+ " that shall be solved by the DSE process. Modules registered with this service"
				+ " are known to the DSE and can be selected by users to define a DSE process.\n"
				+ "The Modules registered here are not defining a a concrete DSE process, just the"
				+ " concrete problems. Exploration modules can be categorized into three classes:"
				+ " Variable definitions, Problem statements and transformations, and Evaluators"
				+ " (solution metrics). The variable definitions consist of a Genotypes"
				+ " (variables), Operators (modify variables), and Creators"
				+ " (variable initialization). Problem statements transform Genotypes into"
				+ " evaluable solutions implemented by Decoders that produce"
				+ " Phenotypes. Evaluators rate Phenotypes such that the quality of a solution can"
				+ " be quantified.\n"
				+ "The complete DSE process is based on DI: each of the elements composing"
				+ " exploration modules have to be implemented as Guice AbstractModules.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Exploration Module Service";
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
